@@ -26,10 +26,15 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
     if (src_body->lib_name) {
         dest_body->lib_name = MVM_malloc(strlen(src_body->lib_name) + 1);
         strcpy(dest_body->lib_name, src_body->lib_name);
-        dest_body->lib_handle = MVM_nativecall_load_lib(dest_body->lib_name);
+        if (src_body->entry_point)
+            dest_body->lib_handle = MVM_nativecall_load_lib(dest_body->lib_name);
     }
 
     /* Rest is just simple copying. */
+    if (src_body->sym_name) {
+        dest_body->sym_name = MVM_malloc(strlen(src_body->sym_name) + 1);
+        strcpy(dest_body->sym_name, src_body->sym_name);
+    }
     dest_body->entry_point = src_body->entry_point;
     dest_body->convention = src_body->convention;
     dest_body->num_args = src_body->num_args;
@@ -37,12 +42,18 @@ static void copy_to(MVMThreadContext *tc, MVMSTable *st, void *src, MVMObject *d
         dest_body->arg_types = MVM_malloc(sizeof(MVMint16) * (src_body->num_args ? src_body->num_args : 1));
         memcpy(dest_body->arg_types, src_body->arg_types, src_body->num_args * sizeof(MVMint16));
     }
-    dest_body->ret_type = src_body->ret_type;
-    if (src_body->jitcode) {
-        dest_body->jitcode = MVM_jit_code_copy(tc, src_body->jitcode);
+    if (src_body->arg_info) {
+        dest_body->arg_info = MVM_malloc(sizeof(MVMObject*) * (src_body->num_args ? src_body->num_args : 1));
+        memcpy(dest_body->arg_info, src_body->arg_info, src_body->num_args * sizeof(MVMObject*));
     }
-    else
-        dest_body->jitcode = NULL;
+    dest_body->ret_type = src_body->ret_type;
+#ifdef HAVE_LIBFFI
+    dest_body->ffi_ret_type = src_body->ffi_ret_type;
+    if (src_body->ffi_arg_types) {
+        dest_body->ffi_arg_types = MVM_malloc(sizeof(ffi_type *) * (dest_body->num_args ? dest_body->num_args : 1));
+        memcpy(dest_body->ffi_arg_types, src_body->ffi_arg_types, sizeof(ffi_type *) * (dest_body->num_args ? dest_body->num_args : 1));
+    }
+#endif
     dest_body->resolve_lib_name = src_body->resolve_lib_name;
     dest_body->resolve_lib_name_arg = src_body->resolve_lib_name_arg;
 }
@@ -73,7 +84,7 @@ static void serialize(MVMThreadContext *tc, MVMSTable *st, void *data, MVMSerial
     MVM_serialization_write_cstr(
         tc,
         writer,
-        !MVM_is_null(tc, body->resolve_lib_name) && !MVM_is_null(tc, body->resolve_lib_name_arg)
+        !body->resolve_lib_name && !MVM_is_null(tc, body->resolve_lib_name_arg)
             ? NULL
             : body->lib_name
     );
@@ -99,13 +110,13 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
     MVMNativeCallBody *body = (MVMNativeCallBody *)data;
     MVMint16 i = 0;
     if (reader->root.version >= 22) {
-        body->lib_name = MVM_serialization_read_cstr(tc, reader);
-        body->sym_name = MVM_serialization_read_cstr(tc, reader);
+        body->lib_name = MVM_serialization_read_cstr(tc, reader, NULL);
+        body->sym_name = MVM_serialization_read_cstr(tc, reader, NULL);
         body->convention = MVM_serialization_read_int(tc, reader);
         body->num_args = MVM_serialization_read_int(tc, reader);
         body->ret_type = MVM_serialization_read_int(tc, reader);
-        body->arg_types = MVM_malloc(body->num_args * sizeof(MVMint16));
-        body->arg_info  = MVM_malloc(body->num_args * sizeof(MVMObject*));
+        body->arg_types = body->num_args ? MVM_malloc(body->num_args * sizeof(MVMint16)) : NULL;
+        body->arg_info  = body->num_args ? MVM_malloc(body->num_args * sizeof(MVMObject*)) : NULL;
         /* TODO ffi support */
         for (i = 0; i < body->num_args; i++) {
             body->arg_types[i] = MVM_serialization_read_int(tc, reader);
@@ -113,7 +124,7 @@ static void deserialize(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, vo
         for (i = 0; i < body->num_args; i++) {
             body->arg_info[i] = MVM_serialization_read_ref(tc, reader);
         }
-        body->resolve_lib_name = MVM_serialization_read_ref(tc, reader);
+        body->resolve_lib_name = (MVMCode *)MVM_serialization_read_ref(tc, reader);
         body->resolve_lib_name_arg = MVM_serialization_read_ref(tc, reader);
 #ifdef HAVE_LIBFFI
         body->ffi_arg_types = MVM_malloc(sizeof(ffi_type *) * (body->num_args ? body->num_args : 1));
@@ -154,8 +165,10 @@ static void gc_cleanup(MVMThreadContext *tc, MVMSTable *st, void *data) {
         MVM_free(body->arg_types);
     if (body->arg_info)
         MVM_free(body->arg_info);
-    if (body->jitcode)
-        MVM_jit_code_destroy(tc, body->jitcode);
+#ifdef HAVE_LIBFFI
+    if (body->ffi_arg_types)
+        MVM_free(body->ffi_arg_types);
+#endif
 }
 
 static void gc_free(MVMThreadContext *tc, MVMObject *obj) {
@@ -175,7 +188,7 @@ const MVMREPROps * MVMNativeCall_initialize(MVMThreadContext *tc) {
 /* gets the setup state */
 static MVMint64 get_int(MVMThreadContext *tc, MVMSTable *st, MVMObject *root, void *data) {
     MVMNativeCallBody *body = (MVMNativeCallBody *)data;
-    return (body->lib_handle ? 1 + (body->jitcode ? 1 : 0) : 0);
+    return (body->lib_handle ? 1 : 0);
 }
 
 static const MVMREPROps NativeCall_this_repr = {
