@@ -9,10 +9,15 @@ static void append_string(char **out, MVMuint32 *size,
     va_list args;
     va_start(args, str);
 
-    vsnprintf(string, line_length, str, args);
+    int append_len = vsnprintf(string, line_length, str, args);
     va_end(args);
+    if (append_len >= line_length) { /* str got truncated by vsnprintf */
+        string[line_length - 1] = '\0';
+        len = line_length;
+    }
+    else
+        len = strlen(string) + 1; /* include terminating \0 so we copy that, too */
 
-    len = strlen(string);
     if (*length + len > *size) {
         while (*length + len > *size)
             *size = *size * 2;
@@ -20,7 +25,7 @@ static void append_string(char **out, MVMuint32 *size,
     }
 
     memcpy(*out + *length, string, len);
-    *length = *length + len;
+    *length = *length + len - 1; /* exclude \0 so we append before it */
 }
 
 static const char * get_typename(MVMuint16 type) {
@@ -73,8 +78,8 @@ static void bytecode_dump_frame_internal(MVMThreadContext *tc, MVMStaticFrame *f
 
     /* mostly stolen from validation.c */
     MVMStaticFrame *static_frame = frame;
-    MVMuint32 bytecode_size = maybe_candidate ? maybe_candidate->bytecode_size : static_frame->body.bytecode_size;
-    MVMuint8 *bytecode_start = maybe_candidate ? maybe_candidate->bytecode : static_frame->body.bytecode;
+    MVMuint32 bytecode_size = maybe_candidate ? maybe_candidate->body.bytecode_size : static_frame->body.bytecode_size;
+    MVMuint8 *bytecode_start = maybe_candidate ? maybe_candidate->body.bytecode : static_frame->body.bytecode;
     MVMuint8 *bytecode_end = bytecode_start + bytecode_size;
     /* current position in the bytestream */
     MVMuint8 *cur_op = bytecode_start;
@@ -100,12 +105,15 @@ static void bytecode_dump_frame_internal(MVMThreadContext *tc, MVMStaticFrame *f
     char mark_this_line = 0;
     MVMCompUnit *cu = static_frame->body.cu;
 
+    /* For handling var-arg ops like the dispatch ops */
+    MVMOpInfo *temporary_op_info = NULL;
+
     while (cur_op < bytecode_end - 1) {
 
         /* allocate a line buffer */
         s = 200;
         l = 0;
-        o = MVM_calloc(s, sizeof(char));
+        o = MVM_malloc(s * sizeof(char));
 
         lineloc = cur_op - bytecode_start;
         /* mark that this line starts at this point in the bytestream */
@@ -159,6 +167,31 @@ static void bytecode_dump_frame_internal(MVMThreadContext *tc, MVMStaticFrame *f
                 op_info = NULL;
             }
         }
+        if (MVM_op_get_mark(op_num)[1] == 'd'
+            || (op_num >= MVM_OP_sp_dispatch_v    && op_num <= MVM_OP_sp_dispatch_o)
+            || (op_num >= MVM_OP_sp_runbytecode_v && op_num <= MVM_OP_sp_runbytecode_o)
+            || (op_num >= MVM_OP_sp_runcfunc_v    && op_num <= MVM_OP_sp_runcfunc_o)
+            || (op_num >= MVM_OP_sp_runnativecall_v && op_num <= MVM_OP_sp_runnativecall_o)
+        ) {
+            /* These are var-arg ops; grab callsite and synthesize op info */
+            MVMCallsite *callsite = MVM_spesh_disp_callsite_for_dispatch_op(op_num, cur_op, cu);
+
+            temporary_op_info = MVM_malloc(MVM_spesh_disp_dispatch_op_info_size(tc, op_info, callsite));
+            MVM_spesh_disp_initialize_dispatch_op_info(tc, op_info, callsite, temporary_op_info);
+
+            op_info = temporary_op_info;
+        }
+
+        if (op_num == MVM_OP_sp_resumption) {
+            MVMSpeshResumeInit *resume_init =
+                &(maybe_candidate->body.resume_inits[GET_I16(cur_op, 2)]);
+            temporary_op_info = MVM_malloc(MVM_spesh_disp_resumption_op_info_size(
+                tc, resume_init->dp, resume_init->res_idx));
+            MVM_spesh_disp_initialize_resumption_op_info(tc,
+                    resume_init->dp, resume_init->res_idx, temporary_op_info);
+
+            op_info = temporary_op_info;
+        }
 
         if (!op_info)
             continue;
@@ -257,8 +290,8 @@ static void bytecode_dump_frame_internal(MVMThreadContext *tc, MVMStaticFrame *f
             }
             else if (op_rw == MVM_operand_read_reg || op_rw == MVM_operand_write_reg) {
                 /* register operand */
-                MVMuint8 frame_has_inlines = maybe_candidate && maybe_candidate->num_inlines ? 1 : 0;
-                MVMuint16 *local_types = frame_has_inlines ? maybe_candidate->local_types : frame->body.local_types;
+                MVMuint8 frame_has_inlines = maybe_candidate && maybe_candidate->body.num_inlines ? 1 : 0;
+                MVMuint16 *local_types = frame_has_inlines ? maybe_candidate->body.local_types : frame->body.local_types;
                 operand_size = 2;
                 a("loc_%u_%s", GET_REG(cur_op, 0),
                     local_types ? get_typename(local_types[GET_REG(cur_op, 0)]) : "unknown");
@@ -295,6 +328,9 @@ static void bytecode_dump_frame_internal(MVMThreadContext *tc, MVMStaticFrame *f
             cur_op += operand_size;
         }
 
+        if (temporary_op_info)
+            MVM_free_null(temporary_op_info);
+
         lines[lineno++] = o;
     }
     {
@@ -370,7 +406,7 @@ char * MVM_bytecode_dump(MVMThreadContext *tc, MVMCompUnit *cu) {
     MVMuint32 s = 1024;
     MVMuint32 l = 0;
     MVMuint32 i, j, k;
-    char *o = MVM_calloc(s, sizeof(char));
+    char *o = MVM_malloc(s * sizeof(char));
     char ***frame_lexicals = MVM_malloc(sizeof(char **) * cu->body.num_frames);
 
     a("\nMoarVM dump of binary compilation unit:\n\n");
@@ -493,7 +529,7 @@ char * MVM_bytecode_dump(MVMThreadContext *tc, MVMCompUnit *cu) {
 void MVM_dump_bytecode_of(MVMThreadContext *tc, MVMFrame *frame, MVMSpeshCandidate *maybe_candidate) {
     MVMuint32 s = 1024;
     MVMuint32 l = 0;
-    char *o = MVM_calloc(s, sizeof(char));
+    char *o = MVM_malloc(s * sizeof(char));
     MVMuint8 *addr;
 
     if (!frame) {
@@ -516,7 +552,7 @@ void MVM_dump_bytecode_of(MVMThreadContext *tc, MVMFrame *frame, MVMSpeshCandida
 void MVM_dump_bytecode_staticframe(MVMThreadContext *tc, MVMStaticFrame *frame) {
     MVMuint32 s = 1024;
     MVMuint32 l = 0;
-    char *o = MVM_calloc(s, sizeof(char));
+    char *o = MVM_malloc(s * sizeof(char));
 
     bytecode_dump_frame_internal(tc, frame, NULL, NULL, NULL, &o, &s, &l);
 
@@ -526,25 +562,29 @@ void MVM_dump_bytecode_staticframe(MVMThreadContext *tc, MVMStaticFrame *frame)
 }
 
 void MVM_dump_bytecode(MVMThreadContext *tc) {
-    MVMStaticFrame *sf = tc->cur_frame->static_info;
-    MVMuint8 *effective_bytecode = MVM_frame_effective_bytecode(tc->cur_frame);
-    if (effective_bytecode == sf->body.bytecode) {
-        MVM_dump_bytecode_of(tc, tc->cur_frame, NULL);
-    } else {
-        MVM_dump_bytecode_of(tc, tc->cur_frame, tc->cur_frame->spesh_cand);
-        /*MVMint32 spesh_cand_idx;*/
-        /*MVMuint8 found = 0;*/
-        /*for (spesh_cand_idx = 0; spesh_cand_idx < sf->body.num_spesh_candidates; spesh_cand_idx++) {*/
+    if (tc->cur_frame != NULL) {
+        MVMStaticFrame *sf = tc->cur_frame->static_info;
+        MVMuint8 *effective_bytecode = MVM_frame_effective_bytecode(tc->cur_frame);
+        if (effective_bytecode == sf->body.bytecode) {
+            MVM_dump_bytecode_of(tc, tc->cur_frame, NULL);
+        } else {
+            MVM_dump_bytecode_of(tc, tc->cur_frame, tc->cur_frame->spesh_cand);
+            /*MVMint32 spesh_cand_idx;*/
+            /*MVMuint8 found = 0;*/
+            /*for (spesh_cand_idx = 0; spesh_cand_idx < sf->body.num_spesh_candidates; spesh_cand_idx++) {*/
             /*MVMSpeshCandidate *cand = sf->body.spesh_candidates[spesh_cand_idx];*/
-            /*if (cand->bytecode == effective_bytecode) {*/
+            /*if (cand->body.bytecode == effective_bytecode) {*/
                 /*MVM_dump_bytecode_of(tc, tc->cur_frame, cand);*/
                 /*found = 1;*/
             /*}*/
-        /*}*/
-        /*if (!found) {*/
+            /*if (!found) {*/
             /* It's likely the MAGIC_BYTECODE from the jit?
              * in that case we just grab tc->cur_frame->spesh_cand apparently */
-        /*}*/
+            /*}*/
+        }
+    }
+    else {
+        fprintf(stderr, "threadcontext has no frame (spesh worker or debug server thread?)\n");
     }
 }
 
@@ -564,7 +604,7 @@ void MVM_dump_bytecode_stackframe(MVMThreadContext *tc, MVMint32 depth) {
         MVMStaticFrameSpesh *spesh = sf->body.spesh;
         for (spesh_cand_idx = 0; spesh_cand_idx < spesh->body.num_spesh_candidates; spesh_cand_idx++) {
             MVMSpeshCandidate *cand = spesh->body.spesh_candidates[spesh_cand_idx];
-            if (cand->bytecode == effective_bytecode) {
+            if (cand->body.bytecode == effective_bytecode) {
                 MVM_dump_bytecode_of(tc, frame, cand);
             }
         }
