@@ -26,9 +26,6 @@ enum {
     MARK_regular  = ' ',
     MARK_special  = '.',
     MARK_sequence = ':',
-    MARK_head     = '+',
-    MARK_body     = '*',
-    MARK_tail     = '-'
 };
 
 typedef struct {
@@ -377,6 +374,28 @@ static void validate_operand(Validator *val, MVMuint32 flags) {
     }
 }
 
+static void validate_dispatch_args(Validator *val, MVMCallsite *cs) {
+    MVMuint16 i;
+    for (i = 0; i < cs->flag_count; i++) {
+        MVMuint8 arg_type = cs->arg_flags[i] & MVM_CALLSITE_ARG_TYPE_MASK;
+        switch (arg_type) {
+            case MVM_CALLSITE_ARG_OBJ:
+                validate_reg_operand(val, MVM_operand_obj);
+                break;
+            case MVM_CALLSITE_ARG_INT:
+                validate_reg_operand(val, MVM_operand_int64);
+                break;
+            case MVM_CALLSITE_ARG_NUM:
+                validate_reg_operand(val, MVM_operand_num64);
+                break;
+            case MVM_CALLSITE_ARG_STR:
+                validate_reg_operand(val, MVM_operand_str);
+                break;
+            default:
+                fail(val, MSG(val, "unrecognized callsite arg type %" PRIu8), arg_type);
+        }
+    }
+}
 
 static void validate_operands(Validator *val) {
     const MVMuint8 *operands = val->cur_info->operands;
@@ -443,6 +462,13 @@ static void validate_operands(Validator *val) {
             else {
                 for (i = 0; i < val->cur_info->num_operands; i++)
                     validate_operand(val, val->cur_info->operands[i]);
+                if (val->cur_mark[1] == 'd') {
+                    /* Dispatch op needs its register args section validated. We
+                     * already just validated the callsite is in range, so can
+                     * fetch it unchecked here. */
+                    MVMCallsite *cs = val->cu->body.callsites[GET_UI16(val->cur_op, -2)];
+                    validate_dispatch_args(val, cs);
+                }
             }
         }
     }
@@ -482,7 +508,6 @@ static void validate_sequence(Validator *val) {
 
             case MARK_regular:
             case MARK_sequence:
-            case MARK_head:
                 unread_op(val);
                 goto terminate_seq;
 
@@ -511,154 +536,6 @@ terminate_seq:
 }
 
 
-static void validate_arg(Validator *val) {
-    MVMCallsiteEntry flags;
-
-    val->remaining_args--;
-
-    if (val->expected_named_arg) {
-        flags = val->expected_named_arg;
-        val->expected_named_arg = 0;
-    }
-    else {
-        MVMuint16 index = val->cur_arg++;
-        MVMuint16 count = val->cur_call->arg_count;
-
-        if (index >= count)
-            fail (val, MSG(val, "argument index %" PRIu16
-                    " not in range 0..%" PRIu32), index, count - 1);
-
-        flags = val->cur_call->arg_flags[index];
-
-        switch (flags & ~MVM_CALLSITE_ARG_MASK) {
-            case 0: /* positionals */
-            case MVM_CALLSITE_ARG_FLAT:
-                val->remaining_positionals--;
-                break;
-
-            case MVM_CALLSITE_ARG_FLAT_NAMED:
-                /* Nothing to do for this case. */
-                break;
-
-            case MVM_CALLSITE_ARG_NAMED:
-                val->expected_named_arg = flags & MVM_CALLSITE_ARG_MASK;
-                goto named_arg;
-
-            default:
-                fail(val, MSG(val, "invalid argument flags (%i) at index %"
-                        PRIu16), (int)(flags & ~MVM_CALLSITE_ARG_MASK), index);
-        }
-    }
-
-    goto regular_arg;
-
-named_arg:
-    if (val->cur_info->opcode != MVM_OP_argconst_s)
-        fail(val, MSG(val, "expected instruction 'argconst_s' but got '%s'"),
-                val->cur_info->name);
-    return;
-
-regular_arg:
-    switch (val->cur_info->opcode) {
-        case MVM_OP_arg_o:
-            if(!(flags & MVM_CALLSITE_ARG_OBJ))
-                goto fail_arg;
-            break;
-
-        case MVM_OP_arg_s:
-            if(!(flags & MVM_CALLSITE_ARG_STR))
-                goto fail_arg;
-            break;
-
-        case MVM_OP_argconst_s:
-            if(!(flags & MVM_CALLSITE_ARG_STR))
-                goto fail_arg;
-            break;
-
-        case MVM_OP_arg_i:
-            if(!(flags & MVM_CALLSITE_ARG_INT))
-                goto fail_arg;
-            break;
-
-        case MVM_OP_arg_n:
-            if(!(flags & MVM_CALLSITE_ARG_NUM))
-                goto fail_arg;
-            break;
-
-        default:
-            fail(val, MSG(val,
-                    "unexpected instruction '%s' during argument preparation"),
-                    val->cur_info->name);
-
-        fail_arg:
-            fail(val, MSG(val, "invalid argument (%i) for instruction %s"),
-                    (int)flags, val->cur_info->name);
-    }
-}
-
-
-static void validate_block(Validator *val) {
-    int block_id = val->cur_mark[1];
-
-    switch (block_id) {
-        case 'a': {
-            MVMuint16 index;
-
-            ensure_op(val, MVM_OP_prepargs);
-            validate_operands(val);
-            index = GET_UI16(val->cur_op, -2);
-            val->cur_call  = val->cu->body.callsites[index];
-            val->cur_arg   = 0;
-            val->expected_named_arg = 0;
-            val->remaining_args = val->cur_call->arg_count;
-            val->remaining_positionals = val->cur_call->num_pos;
-
-            break;
-        }
-
-        default:
-            fail(val, MSG(val, "unknown instruction block '%c'"), block_id);
-    }
-
-    while (val->cur_op < val->bc_end) {
-        int type, id;
-
-        read_op(val);
-        type = val->cur_mark[0];
-        id   = val->cur_mark[1];
-
-        if (val->cur_info->specializable)
-            val->frame->body.specializable = 1;
-
-        if (id != block_id)
-            fail(val, MSG(val, "expected instruction marked '%c' but got '%c'"),
-                    block_id, id);
-
-        switch (type) {
-            case MARK_body: break;
-            case MARK_tail: goto terminate_block;
-            default:        fail_illegal_mark(val);
-        }
-
-        switch (block_id) {
-            case 'a':
-                validate_operands(val);
-                validate_arg(val);
-                break;
-        }
-    }
-
-terminate_block:
-    switch (block_id) {
-        case 'a':
-            validate_operands(val);
-            ensure_no_remaining_positionals(val);
-            ensure_no_remaining_args(val);
-            break;
-    }
-}
-
-
 /* Validate that a static frame's bytecode is executable by the interpreter. */
 void MVM_validate_static_frame(MVMThreadContext *tc,
         MVMStaticFrame *static_frame) {
@@ -716,10 +593,6 @@ void MVM_validate_static_frame(MVMThreadContext *tc,
                 validate_sequence(val);
                 break;
 
-            case MARK_head:
-                validate_block(val);
-                break;
-
             default:
                 fail_illegal_mark(val);
         }
@@ -730,4 +603,7 @@ void MVM_validate_static_frame(MVMThreadContext *tc,
 
     /* Validation successful. Clear up instruction offsets. */
     MVM_free(val->labels);
+
+    /* Mark frame validated. */
+    static_frame->body.validated = 1;
 }
