@@ -355,7 +355,8 @@ void MVM_gc_mark_thread_unblocked(MVMThreadContext *tc) {
                     }
                 }
             } else if (MVM_load(&tc->gc_status) == MVMGCStatus_NONE) {
-                fprintf(stderr, "marking thread %d unblocked, but its status is already NONE.\n", tc->thread_id);
+                if (tc->instance->debugserver && tc->instance->debugserver->debugspam_protocol)
+                    fprintf(stderr, "marking thread %d unblocked, but its status is already NONE.\n", tc->thread_id);
                 break;
             } else {
                 MVM_platform_thread_yield();
@@ -624,13 +625,14 @@ void MVM_gc_enter_from_allocator(MVMThreadContext *tc) {
  * MVMSUSPENDSTATUS_MASK.
  *   */
 static int react_to_debugserver_request(MVMThreadContext *tc) {
-    if (tc->instance->debugserver->debugspam_protocol)
+    if (tc->instance->debugserver && tc->instance->debugserver->debugspam_protocol)
         fprintf(stderr, "thread %p has received a request.\n", tc);
 
     MVMDebugServerRequestKind kind = tc->instance->debugserver->request_data.kind;
 
     if (kind == MVM_DebugRequest_invoke) {
-        MVMObject *invoke_target = tc->instance->debugserver->request_data.data.invoke.target;
+        MVMCode *invoke_target = tc->instance->debugserver->request_data.data.invoke.target;
+        MVMArgs *args = tc->instance->debugserver->request_data.data.invoke.args;
         tc->instance->debugserver->request_data.data.invoke.target = NULL;
 
         /* Have to clear the threadcontext's "blocked" status so that invoke
@@ -641,12 +643,13 @@ static int react_to_debugserver_request(MVMThreadContext *tc) {
             MVM_panic(MVM_exitcode_gcorch, "could not unblock/unsuspend thread");
         }
 
-        STABLE(invoke_target)->invoke(tc, invoke_target, tc->cur_frame->cur_args_callsite, tc->cur_frame->args);
+        MVM_frame_dispatch(tc, invoke_target, *args, -1);
 
         MVM_gc_mark_thread_blocked(tc);
     }
     else {
-        fprintf(stderr, "this debug request kind not implemented: %d\n", kind);
+        if (tc->instance->debugserver && tc->instance->debugserver->debugspam_protocol)
+            fprintf(stderr, "this debug request kind not implemented: %d\n", kind);
         return 0;
     }
 
@@ -654,7 +657,8 @@ static int react_to_debugserver_request(MVMThreadContext *tc) {
                 MVM_DebugRequestStatus_sender_is_waiting,
                 MVM_DebugRequestStatus_receiver_acknowledged)
         != MVM_DebugRequestStatus_sender_is_waiting) {
-        fprintf(stderr, "could not acknowledge request?!?\n");
+        if (tc->instance->debugserver && tc->instance->debugserver->debugspam_protocol)
+            fprintf(stderr, "could not acknowledge request?!?\n");
     }
     tc->instance->debugserver->request_data.kind = MVM_DebugRequest_empty;
 
@@ -669,7 +673,9 @@ void MVM_gc_enter_from_interrupt(MVMThreadContext *tc) {
             fprintf(stderr, "thread %d reacting to suspend request\n", tc->thread_id);
         MVM_gc_mark_thread_blocked(tc);
         while (1) {
+            uv_mutex_lock(&tc->instance->debugserver->mutex_cond);
             uv_cond_wait(&tc->instance->debugserver->tell_threads, &tc->instance->debugserver->mutex_cond);
+            uv_mutex_unlock(&tc->instance->debugserver->mutex_cond);
             if ((MVM_load(&tc->gc_status) & MVMSUSPENDSTATUS_MASK) == MVMSuspendState_NONE) {
                 if (tc->instance->debugserver && tc->instance->debugserver->debugspam_protocol)
                     fprintf(stderr, "thread %d got un-suspended\n", tc->thread_id);
