@@ -8,6 +8,7 @@ static void instrument_graph_with_breakpoints(MVMThreadContext *tc, MVMSpeshGrap
     MVMSpeshBB *bb = g->entry->linear_next;
 
     MVMint32 last_filename = -1;
+    MVMint64 last_line_number = -1;
 
     char *filename_buf = NULL;
 
@@ -26,7 +27,6 @@ static void instrument_graph_with_breakpoints(MVMThreadContext *tc, MVMSpeshGrap
             filename_string_index = bbba->filename_string_heap_index;
             MVM_free(bbba);
         } else {
-            line_number = -1;
             bb = bb->linear_next;
             continue;
         }
@@ -47,7 +47,7 @@ static void instrument_graph_with_breakpoints(MVMThreadContext *tc, MVMSpeshGrap
             continue;
         }
 
-        if (line_number >= 0) {
+        if (line_number >= 0 && line_number != last_line_number) {
             breakpoint_ins = MVM_spesh_alloc(tc, g, sizeof(MVMSpeshIns));
             breakpoint_ins->info        = MVM_op_get_op(MVM_OP_breakpoint);
             breakpoint_ins->operands    = MVM_spesh_alloc(tc, g, 2 * sizeof(MVMSpeshOperand));
@@ -64,6 +64,7 @@ static void instrument_graph_with_breakpoints(MVMThreadContext *tc, MVMSpeshGrap
             breakpoint_ins->operands[1].lit_i32 = line_number;
 
             last_filename = filename_string_index;
+            last_line_number = line_number;
 
             MVM_spesh_manipulate_insert_ins(tc, bb, ins->prev, breakpoint_ins);
         }
@@ -147,7 +148,6 @@ static void instrument_graph(MVMThreadContext *tc, MVMSpeshGraph *g) {
             filename_string_index = bbba->filename_string_heap_index;
             MVM_free(bbba);
         } else {
-            line_number = -1;
             bb = bb->linear_next;
             continue;
         }
@@ -286,6 +286,8 @@ static void line_numbers_instrument(MVMThreadContext *tc, MVMStaticFrame *sf, MV
         if (!sf->body.instrumentation || !sf->body.instrumentation->instrumented_bytecode)
             add_instrumentation(tc, sf, want_coverage);
         sf->body.bytecode      = sf->body.instrumentation->instrumented_bytecode;
+        if (sf->body.handlers)
+            MVM_free(sf->body.handlers);
         sf->body.handlers      = sf->body.instrumentation->instrumented_handlers;
         sf->body.bytecode_size = sf->body.instrumentation->instrumented_bytecode_size;
 
@@ -308,12 +310,11 @@ void MVM_line_coverage_report(MVMThreadContext *tc, MVMString *filename, MVMuint
     if (tc->instance->coverage_control == 2 || (!tc->instance->coverage_control && cache[cache_slot] == 0)) {
         char *encoded_filename;
         char composed_line[256];
-        size_t length;
 
         cache[cache_slot] = 1;
 
         encoded_filename = MVM_string_utf8_encode_C_string(tc, filename);
-        if ((length = snprintf(composed_line, 255, "HIT  %s  %d\n", encoded_filename, line_number)) > 0) {
+        if (snprintf(composed_line, 255, "HIT  %s  %d\n", encoded_filename, line_number) > 0) {
             fputs(composed_line, tc->instance->coverage_log_fh);
         }
         MVM_free(encoded_filename);
