@@ -1,4 +1,5 @@
 #include "moar.h"
+#include "platform/io.h"
 
 #ifndef _WIN32
 #include <sys/types.h>
@@ -164,7 +165,6 @@ void MVM_file_rename(MVMThreadContext *tc, MVMString *src, MVMString *dest) {
 }
 
 void MVM_file_delete(MVMThreadContext *tc, MVMString *f) {
-    uv_fs_t req;
     char * const a = MVM_string_utf8_c8_encode_C_string(tc, f);
 
 #ifdef _WIN32
@@ -176,6 +176,7 @@ void MVM_file_delete(MVMThreadContext *tc, MVMString *f) {
     }
 
 #else
+    uv_fs_t req;
     const int r = uv_fs_unlink(NULL, &req, a, NULL);
 
     if( r < 0 && r != UV_ENOENT) {
@@ -235,25 +236,22 @@ MVMint64 MVM_file_isexecutable(MVMThreadContext *tc, MVMString *filename, MVMint
         return 1;
     else {
         /* true if fileext is in PATHEXT=.COM;.EXE;.BAT;.CMD;.VBS;.VBE;.JS;.JSE;.WSF;.WSH;.MSC */
-        MVMString *dot = MVM_string_ascii_decode_nt(tc, tc->instance->VMString, ".");
-        MVMROOT(tc, dot, {
-            MVMint64 n = MVM_string_index_from_end(tc, filename, dot, 0);
-            if (n >= 0) {
-                MVMString *fileext = MVM_string_substring(tc, filename, n, -1);
-                char *ext  = MVM_string_utf8_c8_encode_C_string(tc, fileext);
-                char *pext = getenv("PATHEXT");
-                int plen   = strlen(pext);
-                int i;
-                for (i = 0; i < plen; i++) {
-                    if (0 == stricmp(ext, pext++)) {
-                         r = 1;
-                         break;
-                    }
+        MVMint64 n = MVM_string_index_from_end(tc, filename, tc->instance->str_consts.dot, 0);
+        if (n >= 0) {
+            MVMString *fileext = MVM_string_substring(tc, filename, n, -1);
+            char *ext  = MVM_string_utf8_c8_encode_C_string(tc, fileext);
+            char *pext = getenv("PATHEXT");
+            int plen   = strlen(pext);
+            int i;
+            for (i = 0; i < plen; i++) {
+                if (0 == stricmp(ext, pext++)) {
+                     r = 1;
+                     break;
                 }
-                MVM_free(ext);
-                MVM_free(pext);
             }
-        });
+            MVM_free(ext);
+            MVM_free(pext);
+        }
     }
     return r;
 }
