@@ -359,6 +359,11 @@ static void proc_async_gc_mark(MVMThreadContext *tc, void *data, MVMGCWorklist *
         MVM_gc_worklist_add(tc, worklist, &(apd->async_task));
 }
 
+static void proc_async_gc_free(MVMThreadContext *tc, MVMObject *root, void *data) {
+    if (data)
+        MVM_free(data);
+}
+
 /* Does an asynchronous close (since it must run on the event loop). */
 static void close_cb(uv_handle_t *handle) {
     MVM_free(handle);
@@ -457,7 +462,7 @@ static const MVMIOOps proc_op_table = {
     NULL,
     NULL,
     proc_async_gc_mark,
-    NULL
+    proc_async_gc_free
 };
 
 static void spawn_async_close(uv_handle_t *handle) {
@@ -944,6 +949,9 @@ static void spawn_gc_mark(MVMThreadContext *tc, void *data, MVMGCWorklist *workl
 static void spawn_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
     if (data) {
         SpawnInfo *si = (SpawnInfo *)data;
+        if (si->prog) {
+            MVM_free_null(si->prog);
+        }
         if (si->cwd) {
             MVM_free_null(si->cwd);
         }
@@ -959,6 +967,12 @@ static void spawn_gc_free(MVMThreadContext *tc, MVMObject *t, void *data) {
                 MVM_free(si->args[i++]);
             MVM_free_null(si->args);
         }
+        if (si->pipe_stdout) {
+            MVM_free_null(si->pipe_stdout);
+        }
+        if (si->pipe_stderr) {
+            MVM_free_null(si->pipe_stderr);
+        }
         MVM_free(si);
     }
 }
@@ -973,12 +987,13 @@ static const MVMAsyncTaskOps spawn_op_table = {
 };
 
 /* Spawn a process asynchronously. */
-MVMObject * MVM_proc_spawn_async(MVMThreadContext *tc, MVMObject *queue, MVMObject *argv,
-                                 MVMString *cwd, MVMObject *env, MVMObject *callbacks) {
+MVMObject * MVM_proc_spawn_async(MVMThreadContext *tc, MVMObject *queue, MVMString *prog,
+                                 MVMObject *argv, MVMString *cwd, MVMObject *env,
+                                 MVMObject *callbacks) {
     MVMAsyncTask  *task;
     MVMOSHandle   *handle;
     SpawnInfo     *si;
-    char          *prog, *_cwd, **_env, **args;
+    char          *_prog, *_cwd, **_env, **args;
     MVMuint64      size, arg_size, i;
     MVMIter       *iter;
     MVMRegister    reg;
@@ -988,17 +1003,20 @@ MVMObject * MVM_proc_spawn_async(MVMThreadContext *tc, MVMObject *queue, MVMObje
         MVM_exception_throw_adhoc(tc,
             "spawnprocasync target queue must have ConcBlockingQueue REPR");
 
-    /* Encode arguments, taking first as program name. */
+    /* Encode program name and arguments. */
     arg_size = MVM_repr_elems(tc, argv);
     if (arg_size < 1)
         MVM_exception_throw_adhoc(tc, "spawnprocasync must have first arg for program");
+
+    _prog = MVM_string_utf8_c8_encode_C_string(tc, prog);
+
+    /* We need a trailing NULL byte, thus the +1. */
     args = MVM_malloc((arg_size + 1) * sizeof(char *));
     for (i = 0; i < arg_size; i++) {
         REPR(argv)->pos_funcs.at_pos(tc, STABLE(argv), argv, OBJECT_BODY(argv), i, &reg, MVM_reg_obj);
         args[i] = MVM_string_utf8_c8_encode_C_string(tc, MVM_repr_get_str(tc, reg.o));
     }
     args[arg_size] = NULL;
-    prog = args[0];
 
     /* Encode CWD. */
     _cwd = MVM_string_utf8_c8_encode_C_string(tc, cwd);
@@ -1025,7 +1043,7 @@ MVMObject * MVM_proc_spawn_async(MVMThreadContext *tc, MVMObject *queue, MVMObje
         MVM_ASSIGN_REF(tc, &(task->common.header), task->body.queue, queue);
         task->body.ops  = &spawn_op_table;
         si              = MVM_calloc(1, sizeof(SpawnInfo));
-        si->prog        = prog;
+        si->prog        = _prog;
         si->cwd         = _cwd;
         si->env         = _env;
         si->args        = args;
@@ -1094,14 +1112,9 @@ void MVM_proc_seed(MVMThreadContext *tc, MVMint64 seed) {
     tinymt64_init(tc->rand_state, (MVMuint64)seed);
 }
 
-/* gets the system time since the epoch truncated to integral seconds */
-MVMint64 MVM_proc_time_i(MVMThreadContext *tc) {
-    return (MVMint64)(MVM_platform_now() / 1000000000);
-}
-
-/* gets the system time since the epoch as floating point seconds */
-MVMnum64 MVM_proc_time_n(MVMThreadContext *tc) {
-    return (MVMnum64)MVM_platform_now() / 1000000000.0;
+/* gets the system time since the epoch in nanoseconds */
+MVMuint64 MVM_proc_time(MVMThreadContext *tc) {
+    return MVM_platform_now();
 }
 
 MVMString * MVM_executable_name(MVMThreadContext *tc) {
@@ -1238,9 +1251,9 @@ MVMint64 MVM_proc_fork(MVMThreadContext *tc) {
     /* Allow MVM_io_eventloop_start to restart the thread if necessary */
     instance->event_loop_thread = NULL;
 
-    MVM_gc_mark_thread_blocked(tc);
+    /* Do not mark thread blocked as the GC also tries to acquire
+     * mutex_threads and it's held only briefly by all holders anyway */
     uv_mutex_lock(&instance->mutex_threads);
-    MVM_gc_mark_thread_unblocked(tc);
 
     /* Check if we are single threaded and if true, fork() */
     if (MVM_thread_cleanup_threads_list(tc, &instance->threads) == 1) {
