@@ -1,21 +1,5 @@
 #include "moar.h"
-
-#ifdef _WIN32
-    #include <winsock2.h>
-    #include <ws2tcpip.h>
-    #include <io.h>
-
-    typedef SOCKET Socket;
-    #define sa_family_t unsigned int
-    #define isatty _isatty
-#else
-    #include "unistd.h"
-    #include <sys/socket.h>
-    #include <sys/un.h>
-
-    typedef int Socket;
-    #define closesocket close
-#endif
+#include "platform/socket.h"
 
 #if defined(_MSC_VER)
 #define snprintf _snprintf
@@ -29,7 +13,9 @@
 /* Error handling varies between POSIX and WinSock. */
 MVM_NO_RETURN static void throw_error(MVMThreadContext *tc, int r, char *operation) MVM_NO_RETURN_ATTRIBUTE;
 #ifdef _WIN32
-    #define MVM_IS_SOCKET_ERROR(x) ((x) == SOCKET_ERROR)
+    #define MVM_IS_INVALID_SOCKET(x) ((x) == INVALID_SOCKET)
+    #define MVM_IS_SOCKET_ERROR(x)   ((x) == SOCKET_ERROR)
+
     static void throw_error(MVMThreadContext *tc, int r, char *operation) {
         int error = WSAGetLastError();
         LPTSTR error_string = NULL;
@@ -41,6 +27,7 @@ MVM_NO_RETURN static void throw_error(MVMThreadContext *tc, int r, char *operati
         MVM_exception_throw_adhoc(tc, "Could not %s: %s", operation, error_string);
     }
 #else
+    #define MVM_IS_INVALID_SOCKET(x) ((x) == -1)
     #define MVM_IS_SOCKET_ERROR(x) ((x) < 0)
     static void throw_error(MVMThreadContext *tc, int r, char *operation) {
         MVM_exception_throw_adhoc(tc, "Could not %s: %s", operation, strerror(errno));
@@ -50,7 +37,7 @@ MVM_NO_RETURN static void throw_error(MVMThreadContext *tc, int r, char *operati
  /* Data that we keep for a socket-based handle. */
 typedef struct {
     /* The socket handle (file descriptor on POSIX, SOCKET on Windows). */
-    Socket handle;
+    MVMSocket handle;
 
     /* Buffer of the last received packet of data, and start/end pointers
      * into the data. */
@@ -74,11 +61,10 @@ static void read_one_packet(MVMThreadContext *tc, MVMIOSyncSocketData *data) {
         MVM_gc_mark_thread_blocked(tc);
         r = recv(data->handle, data->last_packet, PACKET_SIZE, 0);
         MVM_gc_mark_thread_unblocked(tc);
-    } while(r == -1 && errno == EINTR);
+    } while(MVM_IS_SOCKET_ERROR(r) && errno == EINTR);
     MVM_telemetry_interval_stop(tc, interval_id, "syncsocket.read_one_packet");
     if (MVM_IS_SOCKET_ERROR(r) || r == 0) {
-        MVM_free(data->last_packet);
-        data->last_packet = NULL;
+        MVM_free_null(data->last_packet);
         if (r != 0)
             throw_error(tc, r, "receive data from socket");
     }
@@ -88,10 +74,10 @@ static void read_one_packet(MVMThreadContext *tc, MVMIOSyncSocketData *data) {
     }
 }
 
-MVMint64 socket_read_bytes(MVMThreadContext *tc, MVMOSHandle *h, char **buf, MVMint64 bytes) {
+MVMint64 socket_read_bytes(MVMThreadContext *tc, MVMOSHandle *h, char **buf, MVMuint64 bytes) {
     MVMIOSyncSocketData *data = (MVMIOSyncSocketData *)h->body.data;
     char *use_last_packet = NULL;
-    MVMuint16 use_last_packet_start, use_last_packet_end;
+    MVMuint16 use_last_packet_start = 0, use_last_packet_end = 0;
 
     /* If at EOF, nothing more to do. */
     if (data->eof) {
@@ -108,8 +94,7 @@ MVMint64 socket_read_bytes(MVMThreadContext *tc, MVMOSHandle *h, char **buf, MVM
             *buf = MVM_malloc(bytes);
             memcpy(*buf, data->last_packet + data->last_packet_start, bytes);
             if (bytes == last_remaining) {
-                MVM_free(data->last_packet);
-                data->last_packet = NULL;
+                MVM_free_null(data->last_packet);
             }
             else {
                 data->last_packet_start += bytes;
@@ -141,8 +126,7 @@ MVMint64 socket_read_bytes(MVMThreadContext *tc, MVMOSHandle *h, char **buf, MVM
         memcpy(*buf + last_available, data->last_packet, bytes - last_available);
         if (bytes == available) {
             /* We used all of the just-read packet. */
-            MVM_free(data->last_packet);
-            data->last_packet = NULL;
+            MVM_free_null(data->last_packet);
         }
         else {
             /* Still something left in the just-read packet for next time. */
@@ -199,7 +183,7 @@ void socket_truncate(MVMThreadContext *tc, MVMOSHandle *h, MVMint64 bytes) {
 }
 
 /* Writes the specified bytes to the stream. */
-MVMint64 socket_write_bytes(MVMThreadContext *tc, MVMOSHandle *h, char *buf, MVMint64 bytes) {
+MVMint64 socket_write_bytes(MVMThreadContext *tc, MVMOSHandle *h, char *buf, MVMuint64 bytes) {
     MVMIOSyncSocketData *data = (MVMIOSyncSocketData *)h->body.data;
     MVMint64 sent = 0;
     unsigned int interval_id;
@@ -209,8 +193,8 @@ MVMint64 socket_write_bytes(MVMThreadContext *tc, MVMOSHandle *h, char *buf, MVM
     while (bytes > 0) {
         int r;
         do {
-            r = send(data->handle, buf, (int)bytes, 0);
-        } while(r == -1 && errno == EINTR);
+            r = send(data->handle, buf, bytes > 0x40000000 ? 0x40000000 : bytes, 0);
+        } while(MVM_IS_SOCKET_ERROR(r) && errno == EINTR);
         if (MVM_IS_SOCKET_ERROR(r)) {
             MVM_gc_mark_thread_unblocked(tc);
             MVM_telemetry_interval_stop(tc, interval_id, "syncsocket.write_bytes");
@@ -228,7 +212,7 @@ MVMint64 socket_write_bytes(MVMThreadContext *tc, MVMOSHandle *h, char *buf, MVM
 
 static MVMint64 do_close(MVMThreadContext *tc, MVMIOSyncSocketData *data) {
     if (data->handle) {
-        closesocket(data->handle);
+        MVM_platform_close_socket(data->handle);
         data->handle = 0;
     }
     return 0;
@@ -249,7 +233,7 @@ static size_t get_struct_size_for_family(sa_family_t family) {
             return sizeof(struct sockaddr_in6);
         case AF_INET:
             return sizeof(struct sockaddr_in);
-#ifndef _WIN32
+#ifdef MVM_HAS_PF_UNIX
         case AF_UNIX:
             return sizeof(struct sockaddr_un);
 #endif
@@ -259,100 +243,148 @@ static size_t get_struct_size_for_family(sa_family_t family) {
 }
 
 /*
- * This function may return any type of sockaddr e.g. sockaddr_un, sockaddr_in or sockaddr_in6
- * It shouldn't be a problem with general code as long as the port number is kept below the int16 limit: 65536
- * After this it defines the family which may spawn non internet sockaddr's
- * The family can be extracted by (port >> 16) & USHORT_MAX
+ * This function resolves a hostname given a port and family (i.e. domain),
+ * returning a valid address to use with a socket under the given domain,
+ * type, and protocol, in addition to whether or not the socket is passive.
  *
- * Currently supported families:
+ * Valid families:
+ * - SOCKET_FAMILY_INET   (PF_INET)
+ * - SOCKET_FAMILY_INET6  (PF_INET6)
+ * - SOCKET_FAMILY_UNIX   (PF_UNIX)
+ * - SOCKET_FAMILY_UNSPEC (PF_UNSPEC)
  *
- * SOCKET_FAMILY_UNSPEC = 0
- *   Unspecified, in most cases should be equal to AF_INET or AF_INET6
+ * Valid types:
+ * - SOCKET_TYPE_STREAM    (SOCK_STREAM)
+ * - SOCKET_TYPE_DGRAM     (SOCK_DGRAM)
+ * - SOCKET_TYPE_RAW       (SOCK_RAW)
+ * - SOCKET_TYPE_RDM       (SOCK_RDM)
+ * - SOCKET_TYPE_SEQPACKET (SOCK_SEQPACKET)
+ * - SOCKET_TYPE_ANY       (any acceptable type)
  *
- * SOCKET_FAMILY_INET = 1
- *   IPv4 socket
- *
- * SOCKET_FAMILY_INET6 = 2
- *   IPv6 socket
- *
- * SOCKET_FAMILY_UNIX = 3
- *   Unix domain socket, will spawn a sockaddr_un which will use the given host as path
- *   e.g: MVM_io_resolve_host_name(tc, "/run/moarvm.sock", 0, SOCKET_FAMILY_UNIX)
- *   will spawn an unix domain socket on /run/moarvm.sock
+ * Valid protocols:
+ * - SOCKET_PROTOCOL_TCP (IPPROTO_TCP)
+ * - SOCKET_PROTOCOL_UDP (IPPROTO_UDP)
+ * - SOCKET_PROTOCOL_ANY (any acceptable protocol)
  */
 
-struct sockaddr * MVM_io_resolve_host_name(MVMThreadContext *tc, MVMString *host, MVMint64 port, MVMuint16 family) {
-    char *host_cstr = MVM_string_utf8_encode_C_string(tc, host);
-    struct sockaddr *dest;
-    int error;
+struct sockaddr * MVM_io_resolve_host_name(MVMThreadContext *tc,
+        MVMString *host, MVMint64 port,
+        MVMuint16 family, MVMint64 type, MVMint64 protocol,
+        MVMint32 passive) {
+    char *host_cstr     = MVM_string_utf8_encode_C_string(tc, host);
+    char  port_cstr[8];
+
+    struct addrinfo  hints;
     struct addrinfo *result;
-    char port_cstr[8];
-    struct addrinfo hints;
+    struct sockaddr *address;
+    size_t           address_len;
 
-#ifndef _WIN32
-    if (family == SOCKET_FAMILY_UNIX) {
-        struct sockaddr_un *result_un = MVM_malloc(sizeof(struct sockaddr_un));
+    int error;
 
-        MVMuint64 host_len = strlen(host_cstr);
-        if (host_len > 107) {
-            char *waste[] = { host_cstr, NULL };
-            MVM_free(result_un);
-            MVM_exception_throw_adhoc_free(tc, waste, "Socket path '%s' is %"PRIu64" characters, max allowed is 107", host_cstr, host_len);
-        }
-
-        result_un->sun_family = AF_UNIX;
-        strcpy(result_un->sun_path, host_cstr);
-        MVM_free(host_cstr);
-
-        return (struct sockaddr *)result_un;
-    }
-#endif
+    memset(&hints, 0, sizeof(hints));
+    /* XXX: This shouldn't be treating addresses meant for active sockets like
+     *      those for passive ones. */
+    hints.ai_flags = AI_ADDRCONFIG | AI_NUMERICSERV | AI_PASSIVE;
+    /* if (passive) hints.ai_flags |= AI_PASSIVE; */
 
     switch (family) {
-        case SOCKET_FAMILY_UNSPEC:
+        case MVM_SOCKET_FAMILY_UNSPEC:
             hints.ai_family = AF_UNSPEC;
             break;
-        case SOCKET_FAMILY_INET:
+        case MVM_SOCKET_FAMILY_INET:
             hints.ai_family = AF_INET;
             break;
-        case SOCKET_FAMILY_INET6:
+        case MVM_SOCKET_FAMILY_INET6:
             hints.ai_family = AF_INET6;
             break;
-        case SOCKET_FAMILY_UNIX:
-            hints.ai_family = AF_UNIX;
-            break;
+        case MVM_SOCKET_FAMILY_UNIX: {
+#ifdef MVM_HAS_PF_UNIX
+            size_t sun_len = strnlen(host_cstr, MVM_SUN_PATH_SIZE);
+
+            if (sun_len >= MVM_SUN_PATH_SIZE) {
+                char *waste[] = { host_cstr, NULL };
+                MVM_exception_throw_adhoc_free(
+                    tc, waste, "Socket path '%s' is too long (max length supported by this platform is %zu characters)",
+                    host_cstr, MVM_SUN_PATH_SIZE - 1
+                );
+            } else {
+                struct sockaddr_un *result_un = MVM_malloc(sizeof(struct sockaddr_un));
+                result_un->sun_family         = AF_UNIX;
+                strcpy(result_un->sun_path, host_cstr);
+                MVM_free(host_cstr);
+                return (struct sockaddr *)result_un;
+            }
+#else
+            MVM_free(host_cstr);
+            MVM_exception_throw_adhoc(tc, "UNIX sockets are not supported by MoarVM on this platform");
+#endif
+        }
         default:
-            MVM_exception_throw_adhoc(tc, "Unsupported socket family: %hu", family);
+            MVM_free(host_cstr);
+            MVM_exception_throw_adhoc(tc, "Unsupported socket family: %"PRIu16"", family);
             break;
     }
 
-    hints.ai_socktype = 0;
-    hints.ai_flags = AI_ADDRCONFIG | AI_NUMERICSERV | AI_PASSIVE;
-    hints.ai_protocol = 0;
-    hints.ai_addrlen = 0;
-    hints.ai_addr = NULL;
-    hints.ai_canonname = NULL;
-    hints.ai_next = NULL;
+    switch (type) {
+        case MVM_SOCKET_TYPE_ANY:
+            hints.ai_socktype = 0;
+            break;
+        case MVM_SOCKET_TYPE_STREAM:
+            hints.ai_socktype = SOCK_STREAM;
+            break;
+        case MVM_SOCKET_TYPE_DGRAM:
+            hints.ai_socktype = SOCK_DGRAM;
+            break;
+        case MVM_SOCKET_TYPE_RAW:
+            MVM_free(host_cstr);
+            MVM_exception_throw_adhoc(tc, "Support for raw sockets NYI");
+        case MVM_SOCKET_TYPE_RDM:
+            MVM_free(host_cstr);
+            MVM_exception_throw_adhoc(tc, "Support for RDM sockets NYI");
+        case MVM_SOCKET_TYPE_SEQPACKET:
+            MVM_free(host_cstr);
+            MVM_exception_throw_adhoc(tc, "Support for seqpacket sockets NYI");
+        default:
+            MVM_free(host_cstr);
+            MVM_exception_throw_adhoc(tc, "Unsupported socket type: %"PRIi64"", type);
+    }
+
+    switch (protocol) {
+        case MVM_SOCKET_PROTOCOL_ANY:
+            hints.ai_protocol = 0;
+            break;
+        case MVM_SOCKET_PROTOCOL_TCP:
+            hints.ai_protocol = IPPROTO_TCP;
+            break;
+        case MVM_SOCKET_PROTOCOL_UDP:
+            hints.ai_protocol = IPPROTO_UDP;
+            break;
+        default:
+            MVM_free(host_cstr);
+            MVM_exception_throw_adhoc(tc, "Unsupported socket protocol: %"PRIi64"", protocol);
+    }
 
     snprintf(port_cstr, 8, "%d", (int)port);
 
     MVM_gc_mark_thread_blocked(tc);
     error = getaddrinfo(host_cstr, port_cstr, &hints, &result);
     MVM_gc_mark_thread_unblocked(tc);
-    if (error == 0) {
-        size_t size = get_struct_size_for_family(result->ai_addr->sa_family);
-        MVM_free(host_cstr);
-        dest = MVM_malloc(size);
-        memcpy(dest, result->ai_addr, size);
-    }
-    else {
-        char *waste[] = { host_cstr, NULL };
-        MVM_exception_throw_adhoc_free(tc, waste, "Failed to resolve host name '%s' with family %hu. Error: '%s'",
-                                       host_cstr, family, gai_strerror(error));
-    }
-    freeaddrinfo(result);
 
-    return dest;
+    if (error != 0) {
+        char *waste[] = { host_cstr, NULL };
+        MVM_exception_throw_adhoc_free(
+            tc, waste, "Failed to resolve host name '%s' with family %"PRIu16".\nError: %s",
+            host_cstr, family, gai_strerror(error)
+        );
+    }
+
+    MVM_free(host_cstr);
+
+    address_len = get_struct_size_for_family(result->ai_family);
+    address     = MVM_malloc(address_len);
+    memcpy(address, result->ai_addr, address_len);
+    freeaddrinfo(result);
+    return address;
 }
 
 /* Establishes a connection. */
@@ -362,11 +394,11 @@ static void socket_connect(MVMThreadContext *tc, MVMOSHandle *h, MVMString *host
 
     interval_id = MVM_telemetry_interval_start(tc, "syncsocket connect");
     if (!data->handle) {
-        struct sockaddr *dest = MVM_io_resolve_host_name(tc, host, port, family);
+        struct sockaddr *dest = MVM_io_resolve_host_name(tc, host, port, family, MVM_SOCKET_TYPE_STREAM, MVM_SOCKET_PROTOCOL_ANY, 0);
         int r;
 
-        Socket s = socket(dest->sa_family , SOCK_STREAM , 0);
-        if (MVM_IS_SOCKET_ERROR(s)) {
+        MVMSocket s = data->handle = socket(dest->sa_family , SOCK_STREAM , 0);
+        if (MVM_IS_INVALID_SOCKET(s)) {
             MVM_free(dest);
             MVM_telemetry_interval_stop(tc, interval_id, "syncsocket connect");
             throw_error(tc, s, "create socket");
@@ -376,14 +408,12 @@ static void socket_connect(MVMThreadContext *tc, MVMOSHandle *h, MVMString *host
             MVM_gc_mark_thread_blocked(tc);
             r = connect(s, dest, (socklen_t)get_struct_size_for_family(dest->sa_family));
             MVM_gc_mark_thread_unblocked(tc);
-        } while(r == -1 && errno == EINTR);
+        } while(MVM_IS_SOCKET_ERROR(r) && errno == EINTR);
         MVM_free(dest);
         if (MVM_IS_SOCKET_ERROR(r)) {
             MVM_telemetry_interval_stop(tc, interval_id, "syncsocket connect");
             throw_error(tc, s, "connect socket");
         }
-
-        data->handle = s;
     }
     else {
         MVM_telemetry_interval_stop(tc, interval_id, "syncsocket didn't connect");
@@ -394,11 +424,11 @@ static void socket_connect(MVMThreadContext *tc, MVMOSHandle *h, MVMString *host
 static void socket_bind(MVMThreadContext *tc, MVMOSHandle *h, MVMString *host, MVMint64 port, MVMuint16 family, MVMint32 backlog) {
     MVMIOSyncSocketData *data = (MVMIOSyncSocketData *)h->body.data;
     if (!data->handle) {
-        struct sockaddr *dest = MVM_io_resolve_host_name(tc, host, port, family);
+        struct sockaddr *dest = MVM_io_resolve_host_name(tc, host, port, family, MVM_SOCKET_TYPE_STREAM, MVM_SOCKET_PROTOCOL_ANY, 1);
         int r;
 
-        Socket s = socket(dest->sa_family , SOCK_STREAM , 0);
-        if (MVM_IS_SOCKET_ERROR(s)) {
+        MVMSocket s = data->handle = socket(dest->sa_family , SOCK_STREAM , 0);
+        if (MVM_IS_INVALID_SOCKET(s)) {
             MVM_free(dest);
             throw_error(tc, s, "create socket");
         }
@@ -424,8 +454,6 @@ static void socket_bind(MVMThreadContext *tc, MVMOSHandle *h, MVMString *host, M
         r = listen(s, (int)backlog);
         if (MVM_IS_SOCKET_ERROR(r))
             throw_error(tc, s, "start listening on socket");
-
-        data->handle = s;
     }
     else {
         MVM_exception_throw_adhoc(tc, "Socket is already bound or connected");
@@ -459,7 +487,7 @@ MVMint64 socket_getport(MVMThreadContext *tc, MVMOSHandle *h) {
 
 static MVMint64 socket_is_tty(MVMThreadContext *tc, MVMOSHandle *h) {
     MVMIOSyncSocketData *data = (MVMIOSyncSocketData *)h->body.data;
-    return (MVMint64)isatty(data->handle);
+    return (MVMint64)MVM_platform_isatty(data->handle);
 }
 
 static MVMint64 socket_handle(MVMThreadContext *tc, MVMOSHandle *h) {
@@ -502,15 +530,15 @@ static const MVMIOOps op_table = {
 
 static MVMObject * socket_accept(MVMThreadContext *tc, MVMOSHandle *h) {
     MVMIOSyncSocketData *data = (MVMIOSyncSocketData *)h->body.data;
-    Socket s;
+    MVMSocket            s;
 
     unsigned int interval_id = MVM_telemetry_interval_start(tc, "syncsocket accept");
     do {
         MVM_gc_mark_thread_blocked(tc);
         s = accept(data->handle, NULL, NULL);
         MVM_gc_mark_thread_unblocked(tc);
-    } while(s == -1 && errno == EINTR);
-    if (MVM_IS_SOCKET_ERROR(s)) {
+    } while(MVM_IS_INVALID_SOCKET(s) && errno == EINTR);
+    if (MVM_IS_INVALID_SOCKET(s)) {
         MVM_telemetry_interval_stop(tc, interval_id, "syncsocket accept failed");
         throw_error(tc, s, "accept socket connection");
     }
