@@ -93,8 +93,7 @@ static void grow_storage(void *store_ptr, MVMuint64 *num, MVMuint64 *alloc, size
     void **store = (void **)store_ptr;
     if (*num == *alloc) {
         *alloc = *alloc ? 2 * *alloc : 32;
-        *store = MVM_realloc(*store, *alloc * size);
-        memset(((char *)*store) + *num * size, 0, (*alloc - *num) * size);
+        *store = MVM_recalloc(*store, *num * size, *alloc * size);
     }
 }
 
@@ -370,8 +369,13 @@ static void set_static_frame_index(MVMThreadContext *tc, MVMHeapSnapshotState *s
        heap by MVM_cu_string, it will have been allocated in gen2 directly, but not
        marked as live for this gc run. Do it here to prevent it from getting freed
        after taking this heap snapshot, while it's actually still referenced from the
-       comp unit's strings list */
-    if (file_name && file_name->common.header.flags2 & MVM_CF_SECOND_GEN)
+       comp unit's strings list. But only do so if we're actually collecting gen2
+       in this run. Otherwise the flag may still be set during global destruction. */
+    if (
+        tc->instance->gc_full_collect
+        && file_name
+        && file_name->common.header.flags2 & MVM_CF_SECOND_GEN
+    )
         file_name->common.header.flags2 |= MVM_CF_GEN2_LIVE;
 
     MVMuint64 file_idx = get_vm_string_index(tc, ss, file_name);
@@ -510,9 +514,6 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
                 process_collectable(tc, ss, &col, (MVMCollectable *)st, &sc_cache);
                 set_type_index(tc, ss, &col, st);
 
-                MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                    (MVMCollectable *)st->method_cache, "Method cache");
-
                 for (i = 0; i < st->type_check_cache_length; i++)
                     MVM_profile_heap_add_collectable_rel_const_cstr_cached(tc, ss,
                         (MVMCollectable *)st->type_check_cache[i], "Type cache entry", &cache_1);
@@ -527,27 +528,6 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
                         (MVMCollectable *)st->boolification_spec->method,
                         "Boolification method", &cache_2);
 
-                if (st->invocation_spec) {
-                    MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                        (MVMCollectable *)st->invocation_spec->class_handle,
-                        "Invocation spec class handle");
-                    MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                        (MVMCollectable *)st->invocation_spec->attr_name,
-                        "Invocation spec attribute name");
-                    MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                        (MVMCollectable *)st->invocation_spec->invocation_handler,
-                        "Invocation spec invocation handler");
-                    MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                        (MVMCollectable *)st->invocation_spec->md_class_handle,
-                        "Invocation spec class handle (multi)");
-                    MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                        (MVMCollectable *)st->invocation_spec->md_cache_attr_name,
-                        "Invocation spec cache attribute name (multi)");
-                    MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                        (MVMCollectable *)st->invocation_spec->md_valid_attr_name,
-                        "Invocation spec valid attribute name (multi)");
-                }
-
                 MVM_profile_heap_add_collectable_rel_const_cstr_cached(tc, ss,
                     (MVMCollectable *)st->WHO, "WHO", &cache_3);
                 MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
@@ -556,9 +536,6 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
                     (MVMCollectable *)st->HOW, "HOW");
                 MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
                     (MVMCollectable *)st->HOW_sc, "HOW serialization context");
-                MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                    (MVMCollectable *)st->method_cache_sc,
-                    "Method cache serialization context");
 
                 if (st->mode_flags & MVM_PARAMETRIC_TYPE) {
                     MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
@@ -605,9 +582,9 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
                     MVMuint16 *type_map;
                     MVMuint16  name_count = frame->static_info->body.num_lexicals;
                     MVMString **names = frame->static_info->body.lexical_names_list;
-                    if (frame->spesh_cand && frame->spesh_cand->lexical_types) {
-                        type_map = frame->spesh_cand->lexical_types;
-                        count    = frame->spesh_cand->num_lexicals;
+                    if (frame->spesh_cand && frame->spesh_cand->body.lexical_types) {
+                        type_map = frame->spesh_cand->body.lexical_types;
+                        count    = frame->spesh_cand->body.num_lexicals;
                     }
                     else {
                         type_map = frame->static_info->body.lexical_types;
@@ -637,19 +614,6 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
 
                 if (frame->extra) {
                     MVMFrameExtra *e = frame->extra;
-                    if (e->special_return_data && e->mark_special_return_data) {
-                        e->mark_special_return_data(tc, frame, ss->gcwl);
-                        process_gc_worklist(tc, ss, "Special return data");
-                    }
-                    if (e->continuation_tags) {
-                        MVMContinuationTag *tag = e->continuation_tags;
-                        while (tag) {
-                            MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
-                                (MVMCollectable *)tag->tag, "Continuation tag");
-                            col.unmanaged_size += sizeof(MVMContinuationTag);
-                            tag = tag->next;
-                        }
-                    }
                     MVM_profile_heap_add_collectable_rel_const_cstr(tc, ss,
                         (MVMCollectable *)e->dynlex_cache_name,
                         "Dynamic lexical cache name");
@@ -745,13 +709,13 @@ static void process_workitems(MVMThreadContext *tc, MVMHeapSnapshotState *ss) {
  * relation to the current collectable with a constant C string that we
  * should not free. */
 void MVM_profile_heap_add_collectable_rel_const_cstr(MVMThreadContext *tc,
-        MVMHeapSnapshotState *ss, MVMCollectable *collectable, char *desc) {
+        MVMHeapSnapshotState *ss, MVMCollectable *collectable, const char *desc) {
     if (collectable)
         add_reference_const_cstr(tc, ss, desc,
             get_collectable_idx(tc, ss, collectable));
 }
 void MVM_profile_heap_add_collectable_rel_const_cstr_cached(MVMThreadContext *tc,
-        MVMHeapSnapshotState *ss, MVMCollectable *collectable, char *desc, MVMuint64 *cache) {
+        MVMHeapSnapshotState *ss, MVMCollectable *collectable, const char *desc, MVMuint64 *cache) {
     if (collectable)
         add_reference_const_cstr_cached(tc, ss, desc,
             get_collectable_idx(tc, ss, collectable), cache);
@@ -798,11 +762,22 @@ static void record_snapshot(MVMThreadContext *tc, MVMHeapSnapshotCollection *col
 static void destroy_current_heap_snapshot(MVMThreadContext *tc) {
     MVMHeapSnapshotCollection *col = tc->instance->heap_snapshots;
 
+    MVM_free(col->snapshot->stats->type_counts);
+    MVM_free(col->snapshot->stats->type_size_sum);
+    MVM_free(col->snapshot->stats->sf_counts);
+    MVM_free(col->snapshot->stats->sf_size_sum);
+    MVM_free(col->snapshot->stats);
     MVM_free(col->snapshot->collectables);
     MVM_free(col->snapshot->references);
     MVM_free_null(col->snapshot);
 }
 
+static void destroy_toc(MVMHeapDumpTableOfContents *toc) {
+    MVM_free(toc->toc_words);
+    MVM_free(toc->toc_positions);
+    MVM_free(toc);
+}
+
 /* Frees all memory associated with the heap snapshot. */
 static void destroy_heap_snapshot_collection(MVMThreadContext *tc) {
     MVMHeapSnapshotCollection *col = tc->instance->heap_snapshots;
@@ -818,9 +793,9 @@ static void destroy_heap_snapshot_collection(MVMThreadContext *tc) {
     MVM_free(col->static_frames);
 
 #if MVM_HEAPSNAPSHOT_FORMAT == 3
-    MVM_free(col->toplevel_toc);
+    destroy_toc(col->toplevel_toc);
     if (col->second_level_toc)
-        MVM_free(col->second_level_toc);
+        destroy_toc(col->second_level_toc);
 #elif MVM_HEAPSNAPSHOT_FORMAT == 2
     MVM_free(col->index->snapshot_sizes);
     MVM_free(col->index);
@@ -890,8 +865,21 @@ void serialize_attribute_stream(MVMThreadContext *tc, MVMHeapSnapshotCollection
     }
 
     {
-        char namebuf[8] = {0};
-        memcpy(namebuf, name, 8);
+        char namebuf[8];
+        /* Yes, this is a lot of boiler plate to silence a bogus gcc warning (but not clang) :(
+         * Unfortunately, this is a real edge case. Using memcpy would lead to
+         * a buffer overflow if name is shorter than 8 bytes. The gcc compiler
+         * warning aside, strncpy seems like exactly the right tool for the job
+         * as we want at most 8 bytes and don't care for any trailing \0, but
+         * are OK with zero padding at the end. */
+#if defined(__GNUC__) && !defined(__clang__)
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wstringop-truncation"
+#endif
+        strncpy(namebuf, name, 8);
+#if defined(__GNUC__) && !defined(__clang__)
+#pragma GCC diagnostic pop
+#endif
         fwrite(namebuf, 8, 1, fh);
     }
 
@@ -978,8 +966,8 @@ void string_heap_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCollect
     char typename[8] = "strings\0";
     MVMuint64 size = 0;
 
-    MVMuint64 size_position = 0; 
-    MVMuint64 end_position = 0; 
+    MVMuint64 size_position = 0;
+    MVMuint64 end_position = 0;
 
     while (i < col->num_strings) {
         char *str = col->strings[i];
@@ -1001,8 +989,10 @@ void string_heap_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCollect
         i++;
     }
 
-    if (result_buffer_insert_pos == result_buffer)
+    if (result_buffer_insert_pos == result_buffer) {
+        MVM_free(result_buffer_insert_pos);
         return;
+    }
 
     size_position = ftell(fh);
     fwrite(&typename, sizeof(char), 8, fh);
@@ -1012,6 +1002,7 @@ void string_heap_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCollect
     cstream = ZSTD_createCStream();
 
     if (ZSTD_isError(return_value = ZSTD_initCStream(cstream, ZSTD_COMPRESSION_VALUE))) {
+        MVM_free(result_buffer);
         MVM_panic(1, "ZSTD compression error in heap snapshot: %s", ZSTD_getErrorName(return_value));
     }
 
@@ -1033,6 +1024,8 @@ void string_heap_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCollect
         return_value = ZSTD_compressStream(cstream, &outbuf, &inbuf);
 
         if (ZSTD_isError(return_value)) {
+            MVM_free(outbuf.dst);
+            MVM_free(result_buffer);
             MVM_panic(1, "Error compressing heap snapshot data: %s", ZSTD_getErrorName(return_value));
         }
 
@@ -1059,6 +1052,8 @@ void string_heap_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCollect
 
 
     if (ZSTD_isError(return_value)) {
+        MVM_free(outbuf.dst);
+        MVM_free(result_buffer);
         MVM_panic(1, "Error compressing heap snapshot data: %s", ZSTD_getErrorName(return_value));
     }
 
@@ -1638,7 +1633,7 @@ void collectables_to_filehandle_ver2(MVMThreadContext *tc, MVMHeapSnapshotCollec
 
     entry->collectables_size += s->num_collectables * i + 4 + sizeof(MVMuint64) * 2;
 
-    
+
 #if DUMP_EVERYTHING_RAW
     gzFile kind_fh, tofi_fh, collsize_fh, unmansize_fh, refstart_fh, refcount_fh;
     kind_fh = open_coll_file(col, "kind");
@@ -1757,6 +1752,7 @@ static void filemeta_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCol
     fwrite(&size, sizeof(MVMuint64), 1, fh);
 
     fputs(metadata, fh);
+    MVM_free(metadata);
     fputc(0, fh);
 
     end_position = ftell(fh);
@@ -1808,6 +1804,7 @@ static void snapmeta_to_filehandle_ver3(MVMThreadContext *tc, MVMHeapSnapshotCol
     fwrite(&size, sizeof(MVMuint64), 1, fh);
 
     fputs(metadata, fh);
+    MVM_free(metadata);
     fputc(0, fh);
 
     end_position = ftell(fh);
